@@ -55,14 +55,17 @@
 #define AVALON_MAX_MINER_NUM 0x100
 #define AVALON_DEFAULT_ASIC_NUM 0xA
 
+/* Default number of miners for Bitburner Fury is for a stack of 8 boards,
+   but it will work acceptably for smaller stacks, too */
+#define BITBURNER_FURY_DEFAULT_MINER_NUM 128
+#define BITBURNER_FURY_DEFAULT_FREQUENCY 256
+#define BITBURNER_FURY_DEFAULT_TIMEOUT 50
+
 #define AVALON_AUTO_CYCLE 1024
 
 #define AVALON_FTDI_READSIZE 510
-#define AVALON_USB_PACKETSIZE 512
 #define AVALON_READBUF_SIZE 8192
-#define AVALON_RESET_TIMEOUT 100
-#define AVALON_READ_TIMEOUT 18 /* Enough to only half fill the buffer */
-#define AVALON_LATENCY 1
+#define AVALON_LATENCY 16 /* This is basically the ftdi default anyway */
 
 struct avalon_task {
 	uint8_t reset		:1;
@@ -140,9 +143,11 @@ struct avalon_info {
 	pthread_t write_thr;
 	pthread_mutex_t lock;
 	pthread_mutex_t qlock;
-	pthread_cond_t qcond;
-	int nonces;
+	cgsem_t qsem;
+	cgtimer_t cgsent;
+	int send_delay;
 
+	int nonces;
 	int auto_queued;
 	int auto_nonces;
 	int auto_hw;
