@@ -2385,7 +2385,7 @@ function ss($a, $b)
  $la = strlen($a);
  $lb = strlen($b);
  if ($la != $lb)
-	return $la - $lb;
+	return $lb - $la;
  return strcmp($a, $b);
 }
 #
